/*
 * text-mmap.c -- mmap a text file into memory,
 *                ensuring that the text is NUL-byte terminated.

   Copyright (C) 2006 Free Software Foundation, Inc.
   Written by Bruno Haible <bruno@clisp.org>, 2003.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <config.h>
#include "text-mmap.h"

/* * * * * */

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#if HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif

#if HAVE_FCNTL_H
# include <fcntl.h>
#endif

/* * * * * */

#if HAVE_UNISTD_H
# include <unistd.h>
#else
  extern int getpagesize();
#endif

#if HAVE_SYSCONF
# ifndef  _SC_PAGE_SIZE
#  define _SC_PAGESIZE _SC_PAGE_SIZE
# endif
# define GETPAGESIZE() sysconf(_SC_PAGESIZE)

#elif HAVE_GETPAGESIZE()
# define GETPAGESIZE() getpagesize()

#else
# define GETPAGESIZE() 0x400 /* 1K ?? or just quit? */
#endif

/* * * * * */

#if HAVE_ERRNO_H
# include <errno.h>
#else
  extern int errno;
#endif

#ifndef ENOMEM
# define ENOMEM 12
#endif

#ifndef EINVAL
# define EINVAL 22
#endif

/* * * * * */

/*=export_func  text_mmap
 *
 * ifdef: HAVE_MMAP
 *
 * what:  map a text file with terminating NUL
 *
 * arg:   char const *,   file,    name of the file to map
 * arg:   int,            prot,    mmap protections (see mmap(2))
 * arg:   int,            flags,   mmap flags (see mmap(2))
 * arg:   txtmap_info_t*, mapinfo, returned info about the mapping
 *
 * ret-type:   void*
 * ret-desc:   The mmaped data address
 *
 * doc:
 *
 * This routine will mmap a file into memory ensuring that there is at least
 * one @file{NUL} character following the file data.  It will return the
 * address where the file contents have been mapped into memory.  If there is a
 * problem, then it will return @code{MAP_FAILED} and set @file{errno}
 * appropriately.
 *
 * The named file does not exist, @code{stat(2)} will set @file{errno} as it
 * will.  If the file is not a regular file, @file{errno} will be
 * @code{EINVAL}.  At that point, @code{open(2)} is attempted with the access
 * bits set appropriately for the requested @code{mmap(2)} protections and flag
 * bits.  On failure, @file{errno} will be set according to the documentation
 * for @code{open(2)}.  If @code{mmap(2)} fails, @file{errno} will be set as
 * that routine sets it.  If @code{txt_mmap} works to this point, a valid
 * address will be returned, but there may still be ``issues''.
 *
 * If the file size is not an even multiple of the system page size, then
 * @code{text_map} will return at this point and @file{errno} will be zero.
 * Otherwise, an anonymous map is attempted.  If not available, then an attempt
 * is made to @code{mmap(2)} @file{/dev/zero}.  If any of these fail, the
 * address of the file's data is returned, bug @code{no} @file{NUL} characters
 * are mapped after the end of the data.
 *
 * see: mmap(2), open(2), stat(2)
 *
 * err: Any error code issued by mmap(2), open(2), stat(2) is possible.
 *      Additionally, if the specified file is not a regular file, then
 *      errno will be set to @code{EINVAL}.
 *
 * example:
 * #include <text-mmap.h>
 * txtmap_info_t mi;
 * int no_nul;
 * void* data = text_mmap( "file", PROT_WRITE, MAP_PRIVATE, &mi );
 * if (data == MAP_FAILED) return;
 * no_nul = (mi.txt_size == mi.txt_full_size);
 * << use the data >>
 * text_munmap( &mi );
=*/
void*
text_mmap (char const * fname, int prot, int flags, txtmap_info_t* txtmi)
{
  /*
   *  Make sure we can stat the regular file.  Save the file size.
   */
  {
    struct stat sb;
    if (stat (fname, &sb) != 0) {
      txtmi->txt_errno = errno;
      return MAP_FAILED;
    }

    if (! S_ISREG(sb.st_mode)) {
      txtmi->txt_errno = errno = EINVAL;
      return MAP_FAILED;
    }

    txtmi->txt_size = sb.st_size;
  }

  /*
   *  Map mmap flags and protections into open flags and do the open.
   */
  {
    int o_flag;
    /*
     *  See if we will be updating the file
     */
    if (prot & PROT_WRITE) {
      o_flag = O_RDWR;

      /*
       *  If you're not sharing the file
       *  then don't let anyone else have access to the file.
       */
      if ((flags & MAP_SHARED) == 0)
        o_flag |= O_EXCL;
    }

    else
      o_flag = O_RDONLY;

    txtmi->txt_fd = open (fname, o_flag);
  }

  if (txtmi->txt_fd < 0) {
    txtmi->txt_errno = errno;
    return MAP_FAILED;
  }

#if HAVE_MMAP
  /*
   *  do the mmap.  If we fail, then preserve errno, close the file and
   *  return the failure.
   */
  txtmi->txt_data = mmap (NULL, txtmi->txt_size, prot, flags, txtmi->txt_fd, 0);
  if (txtmi->txt_data == MAP_FAILED) {
    txtmi->txt_errno = errno;
    close (txtmi->txt_fd);
    txtmi->txt_fd = -1;
    errno = txtmi->txt_errno;
    return MAP_FAILED;
  }

  /*
   *  Most likely, everything will turn out fine now.  The only difficult
   *  part at this point is coping with files with sizes that are a multiple
   *  of the page size.  Handling that is what this whole thing is about.
   */
  txtmi->txt_zero_fd = -1;
  txtmi->txt_errno   = 0;

  do {
    txt_size_t pgsz = GETPAGESIZE();

    /*
     *  Compute the pagesize rounded mapped memory size.
     *  IF this is not the same as the file size, then there are NUL's
     *  at the end of the file mapping and all is okay.
     */
    txtmi->txt_full_size = (txtmi->txt_size + (pgsz - 1)) & ~(pgsz - 1);
    if (txtmi->txt_size != txtmi->txt_full_size)
      break;

    /*
     *  Still here?  We have to append a page of NUL's
     */
    txtmi->txt_full_size += pgsz;
    {
      void* pNuls;

#ifdef MAP_ANONYMOUS
      pNuls = mmap ((void*)(((char*)txtmi->txt_data) + txtmi->txt_size), pgsz,
                    PROT_READ, MAP_SHARED|MAP_ANONYMOUS|MAP_FIXED, 0, 0);
#else
      txtmi->txt_zero_fd = open ("/dev/zero", O_RDONLY);
      if (txtmi->txt_zero_fd < 0)
        pNuls = MAP_FAILED;

      else pNuls =
             mmap ((void*)(((char*)txtmi->txt_data) + txtmi->txt_size), pgsz,
                   PROT_READ, MAP_PRIVATE|MAP_FIXED, txtmi->txt_zero_fd, 0);
#endif

      /*
       *  If we couldn't map the page of NULs, then the caller can
       *  figure this out by checking that full_size == size and
       *  the txt_errno is not zero.
       */
      if (pNuls == MAP_FAILED) {
        txtmi->txt_errno = errno;
        txtmi->txt_full_size = txtmi->txt_size;
      }
    }
  } while(0);

#else
  /*
   *  This platform does not have mmap, so we open/malloc/read/close & free
   *  the data instead.
   */
  txtmi->txt_data = malloc (txtmi->txt_size + 1);
  if (txtmi->txt_data == NULL) {
    close (txtmi->txt_fd);
    txtmi->txt_fd = -1;
    errno = ENOMEM;
    return MAP_FAILED;
  }

  {
    txt_size_t rem = txtmi->txt_size;
    char *     buf = txtmi->txt_data;

    /*
     *  IF there are any data, then slurp it all in.
     *  Choke if we don't get it all.
     */
    if (rem > 0) for (;;) {
      txt_ssize_t rct = read (txtmi->txt_fd, buf, rem);
      if (rct < 0) {
        txtmi->txt_errno = errno;
        free (txtmi->txt_data);
        close (txtmi->txt_fd);
        return MAP_FAILED;
      }
      buf += rct;
      rem -= rct;
      if (rem == 0)
        break;
    }

    *buf = '\0';
  }
  close (txtmi->txt_fd);

#endif
  return txtmi->txt_data;
}


/*=export_func  txt_munmap
 *
 * what:  unmap the data mapped in by txt_mmap
 *
 * arg:   txtmap_info_t*, mapinfo, info about the mapping
 *
 * ret-type:   int
 * ret-desc:   -1 or 0.  @file{errno} will have the error code.
 *
 * doc:
 *
 * This routine will unmap the data mapped in with @code{txt_mmap} and close
 * the associated file descriptors opened by that function.
 *
 * see: munmap(2), close(2)
 *
 * err: Any error code issued by munmap(2) or close(2) is possible.
=*/
int
text_munmap (txtmap_info_t* txtmi)
{
#if HAVE_MMAP
  int res = munmap (txtmi->txt_data, txtmi->txt_full_size);
  if (res != 0)
    goto error_return;

  res = close (txtmi->txt_fd);
  if (res != 0)
    goto error_return;

  txtmi->txt_fd = -1;
  errno = 0;
  if (txtmi->txt_zero_fd != -1) {
    res = close (txtmi->txt_zero_fd);
    if (res == 0)
      txtmi->txt_zero_fd = -1;
  }

 error_return:
  txtmi->txt_errno = errno;
  return res;
#else
  free (txtmi->txt_data);
  return 0;
#endif
}
/*
 * Local Variables:
 * c-file-style: "gnu"
 * indent-tabs-mode: nil
 * End:
 * text_mmap ends here */
