/*
 * text-mmap.h -- mmap a text file into memory,
 *                ensuring that the text is NUL-byte terminated.
 *
 * mmap(2) assumes that the subject file may be dynamically resized while
 * the mapping is active.  Consequently, if you mmap a text file that
 * happens to have a size that is a multiple of the system page size,
 * there will be *no* NUL byte following the end of the data.  Consequently,
 * strlen(3C) and friends will seg fault if you use them.  This module
 * will ensure that the string-terminating NUL byte is present, but it
 * also assumes that the file size is static for the duration of the
 * mapping.  If that may not be a valid assumption, the programmer is
 * encouraged to not use this function.  Use a open/malloc/read/close/free
 * sequence instead.

   Copyright (C) 2006 Free Software Foundation, Inc.
   Written by Bruce Korb <bkorb@gnu.org>, 2000

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef _TEXT_MMAP_H_
#define _TEXT_MMAP_H_ 1

#if HAVE_STDDEF_H
# include <stddef.h>
# define  txt_size_t  size_t
# define  txt_ssize_t ssize_t
#else
  typedef unsigned long txt_size_t;
  typedef long txt_ssize_t;
#endif

#if HAVE_SYS_MMAN_H
#  include <sys/mman.h>
#endif

/*
 *  txt_mmap structure.  Only active on platforms with mmap(2).
 */
#ifndef MAP_FAILED
#  define  MAP_FAILED  ((void*)-1)
#endif

/*
 *  This is an output only structure used by txt_mmap and txt_munmap.
 *  Clients must not alter the contents and must provide it to both
 *  the txt_mmap and txt_munmap procedures.  BE ADVISED: if you are
 *  mapping the file with PROT_WRITE the NUL byte at the end MIGHT NOT
 *  BE WRITABLE.  In any event, that byte is not be written back
 *  to the source file.
 */
typedef struct {
    void*       txt_data;      /* text file data   */
    txt_size_t  txt_size;      /* actual file size */
    int         txt_fd;        /* file descriptor  */
    txt_size_t  txt_full_size; /* mmaped mem size  */
    int         txt_zero_fd;   /* fd for /dev/zero */
    int         txt_errno;     /* warning code     */
} txtmap_info_t;

extern void*
text_mmap (char const * file, int prot, int flags, txtmap_info_t* mi);

extern int
txt_munmap (txtmap_info_t* txtmi);

#endif /* _TEXT_MMAP_H_ */
/*
 * Local Variables:
 * c-file-style: "gnu"
 * indent-tabs-mode: nil
 * End:
 * text_mmap ends here */
